import processing.core.*; 
import processing.xml.*; 

import peasy.*; 
import processing.opengl.*; 

import java.applet.*; 
import java.awt.Dimension; 
import java.awt.Frame; 
import java.awt.event.MouseEvent; 
import java.awt.event.KeyEvent; 
import java.awt.event.FocusEvent; 
import java.awt.Image; 
import java.io.*; 
import java.net.*; 
import java.text.*; 
import java.util.*; 
import java.util.zip.*; 
import java.util.regex.*; 

public class Balloon extends PApplet {


//import net.java.games.jogl.*;


PeasyCam cam;
// GL gl;

float axis_X = 0;
float axis_Y = 0;
float axis_Z = 0;

float control_X = 100;
float control_Y = 100;
float control_Z = 0;

float offset_X = 0;
float offset_Y = 100;
float offset_Z = 0;

int totalSegments = 10;

PVector[] orbit = new PVector[totalSegments];
Segment[] segments = new Segment[totalSegments];

PVector segmentOffset;

public void setup() {
	size(1200,800,P3D); 
	// gl = ((PGraphicsGL)g).gl;
	// gl.glEnable(GL.GL_CULL_FACE);

	cam = new PeasyCam(this, 1000);
	cam.setMinimumDistance(50);
	cam.setMaximumDistance(2000);

	
    // --------------------------------------------------------------------
    // parse data
    // --------------------------------------------------------------------

	for(int i=0;i<totalSegments;i++){
		segments[i] = new Segment();
		if(i==0){
			segments[i].previousSegment = segments[0];
		} else {
			segments[i].previousSegment = segments[i-1];
		}
		
		segments[i].seed = i;
	}

}
public void draw() {
	
	background(0);
	
	drawAxes(400);
	
	// --------------------------------------------------------------------
	// draw base 
	// --------------------------------------------------------------------
	float angleX,angleY;
	int i,j,k;
	float percentage;

	segmentOffset = new PVector(0,0,0);
	// --------------------------------------------------------------------
	// parse data
	// --------------------------------------------------------------------
	
	// custom position of segment 2
	// to see values change easily
	
	
	
	segments[1].endRelativeVec.y = .5f*(mouseY-height/2);
	segments[1].endRelativeVec.x = .5f*(mouseX-width/2);
	segments[1].resetRelatives();
	
	segments[3].endRelativeVec.y = -.25f*(mouseY-height/2);
	segments[3].endRelativeVec.x = -.25f*(mouseX-width/2);
	segments[3].resetRelatives();
	
	
	for(i=0;i<totalSegments;i++){
		calculateSegmentPosition(i);
		segments[i].resetRelatives();

	}
	


	i = 0;
	segments[i].setPreviousOrbit(segments[i].startOrbit);
	segments[i].setNextOrbit(segments[i+1].startOrbit);
	segments[i].parseLerps();
	
	for(i=1;i<(totalSegments-1);i++){
		segments[i].setPreviousOrbit(segments[i-1].endOrbit);
		segments[i].setNextOrbit(segments[i+1].startOrbit);
		segments[i].parseLerps();
	}

	i = totalSegments-1;
	segments[i].setPreviousOrbit(segments[i-1].endOrbit);
	segments[i].setNextOrbit(segments[i].endOrbit);
	segments[i].parseLerps();
	
	
	for(i=0;i<totalSegments;i++){
		segments[i].draw();
		// segments[i].drawWireframes();
		// segments[i].drawGuides();
	}
}


public void calculateSegmentPosition(int i){

	segments[i].createAbsolutes();
	segments[i].updateAbsolutes(segmentOffset);
	segmentOffset.add(segments[i].endRelativeVec.get());
	segments[i].parse();
}



public void keyPressed() {
	if (key == 'r' || key == 'R') {
		for(int i=0;i<totalSegments;i++){
			segments[i].createRelatives();
			segments[i].resetRelatives();
			segments[i].createAbsolutes();
		}
	}
}

public void drawAxes(float radius) {
  noFill();
  strokeWeight(1);
  stroke(255, 0, 0,255);
  line(0, 0, 0, radius, 0, 0);
  stroke(0, 255, 0,255);
  line(0, 0, 0, 0, radius, 0);
  stroke(0, 0, 255,255);
  line(0, 0, 0, 0, 0, radius);

  stroke(0xff00FF00,51);
  pushMatrix();
  translate(0, radius,0);
  box(10);
  popMatrix();             

  stroke(0xffFF0000,51);
  pushMatrix();
  translate(radius, 0,0);
  box(10);
  popMatrix(); 

  stroke(0xff0000FF,51);
  pushMatrix();
  translate(0, 0,radius);
  box(10);
  popMatrix(); 

  stroke(0xffFFFFFF,51);
  ellipse(0, 0, 200, 200);

  stroke(0xffFFFFFF,51);
  box(radius*2);
}

public Quaternion quat_mult(Quaternion q1, Quaternion q2) {

  float a = (q1.a*q2.a - q1.b*q2.b - q1.c*q2.c - q1.d*q2.d);
  float b = (q1.a*q2.b + q1.b*q2.a + q1.c*q2.d - q1.d*q2.c);
  float c = (q1.a*q2.c - q1.b*q2.d + q1.c*q2.a + q1.d*q2.b);
  float d = (q1.a*q2.d + q1.b*q2.c - q1.c*q2.b + q1.d*q2.a);
  Quaternion result = new Quaternion(a,b,c,d);
  return result;
}

// Quaternion multiplication without the .a component
public PVector quat_pointmult( Quaternion q1, Quaternion q2) {
  float x = (q1.a*q2.b + q1.b*q2.a + q1.c*q2.d - q1.d*q2.c);
  float y = (q1.a*q2.c - q1.b*q2.d + q1.c*q2.a + q1.d*q2.b);
  float z = (q1.a*q2.d + q1.b*q2.c - q1.c*q2.b + q1.d*q2.a);
  PVector result = new PVector(x,y,z);
  return result;
}

class Quaternion {
  public float a = 0;
  public float b = 0;
  public float c = 0;
  public float d = 1;

  public Quaternion(float aa,float bb,float cc,float dd) {
    this.a = aa;
    this.b = bb;
    this.c = cc;
    this.d = dd;
  }
}
public PVector rotateAroundArbitraryAxis(PVector offsetParam,PVector controlParam,PVector axisParam,float angle) {
	PVector control = controlParam.get();
	PVector axis = axisParam.get();
	PVector offset = offsetParam.get();

	control.sub(offset);
	axis.sub(offset);
	axis.normalize(axis);

	float rads = PI/180 * angle;                       
	float sinTheta = sin(rads*.5f);
	float cosTheta = cos(rads*.5f);

	Quaternion rotQuat   = new Quaternion(  cosTheta,  
	sinTheta * axis.x,  
	sinTheta * axis.y,  
	sinTheta * axis.z );
	Quaternion pointQuat = new Quaternion(  0,
	control.x,
	control.y,
	control.z );
	Quaternion conjQuat  = new Quaternion(  rotQuat.a,
	-rotQuat.b,
	-rotQuat.c,
	-rotQuat.d );

	Quaternion temp = quat_mult(rotQuat,pointQuat);
	PVector result = quat_pointmult(temp,conjQuat);
	result.add(offset);
	return result;
}

public PVector rotateAroundZAxis(PVector offset, PVector controlParam, float angle) {

	PVector control = controlParam.get();
	control.sub(offset);

	float rads = PI/180 * angle;
	float cosTheta = cos(rads);
	float sinTheta = sin(rads);

	float x = cosTheta*control.x + sinTheta*control.y;
	float y = -sinTheta*control.x + cosTheta*control.y;
	float z = control.z;

	PVector result = new PVector(x,y,z);
	result.add(offset);

	return result;
}

public PVector rotateAroundYAxis(PVector offset, PVector controlParam, float angle) {
	PVector control = controlParam.get();
	control.sub(offset);

	float rads = PI/180 * angle;
	float cosTheta = cos(rads);
	float sinTheta = sin(rads);

	float x = cosTheta*control.x - sinTheta*control.z;
	float y = control.y;
	float z = sinTheta*control.x + cosTheta*control.z;

	PVector result = new PVector(x,y,z);

	result.add(offset);

	return result;
}

public PVector rotateAroundXAxis(PVector offset, PVector controlParam, float angle) {
	PVector control = controlParam.get();
	control.sub(offset);

	float rads = PI/180 * angle;
	float cosTheta = cos(rads);
	float sinTheta = sin(rads);

	float x = control.x;
	float y = cosTheta*control.y + sinTheta*control.z;
	float z = -sinTheta*control.y + cosTheta*control.z;

	PVector result = new PVector(x,y,z);

	result.add(offset);

	return result;
}


class Segment {

	public PVector offset;
	public PVector axis;
	public PVector control;


	public PVector storedOffset;
	public PVector storedAxis;
	public PVector storedControl;

	public int seed;

	public float angle;

	public int totalLines = 50;

	public PVector[] baseOrbit = new PVector[totalLines];
	public PVector[] startOrbit = new PVector[totalLines];
	public PVector[] startMarginOrbit = new PVector[totalLines];
	public PVector[] endOrbit = new PVector[totalLines];
	public PVector[] endMarginOrbit = new PVector[totalLines];
	public PVector[] midOrbit = new PVector[totalLines];

	public PVector[] prevStartOrbit = new PVector[totalLines];
	public PVector[] nextStartOrbit = new PVector[totalLines];
	
	public PVector[] startLerpOrbit = new PVector[totalLines];
	public PVector[] endLerpOrbit = new PVector[totalLines];

	
	public PVector zeroVect;

	public PVector startRelativeVec;
	public PVector endRelativeVec;
	public PVector startCrossRelativeVec;
	public PVector endCrossRelativeVec;
	public PVector pseudoRelativeVec;

	public PVector startAbsoluteVec;
	public PVector endAbsoluteVec;
	public PVector startCrossAbsoluteVec;
	public PVector endCrossAbsoluteVec;
	public PVector pseudoAbsoluteVec;

	public Segment previousSegment;

	int from = color(255, 0, 102);
	int to = color(0, 102, 153);

	public float scalar = random(50) + 50;

	public Segment() {
		int radius = 25;
		zeroVect	= new PVector( 0,0,0);
		
		axis 	= new PVector( 25, 0, 0 );
		offset 	= new PVector( 0,0,0);
		control = new PVector( 0, 10, 0 );
		
		storedAxis = new PVector( 0, random(25), random(25) );
		storedOffset = offset.get();
		storedControl = control.get();
		
		createRelatives();
		resetRelatives();
		createAbsolutes();
	}

	public void createRelatives(){
		int radius = 300;
		float half_radius = radius*.5f;
		startRelativeVec = new PVector( 0, 0, 0 );
		endRelativeVec = new PVector( random(radius)-half_radius, random(radius)-half_radius, random(radius)-half_radius );
		
	}

	public void resetRelatives(){
		
		// scalar = mouseY-height/2;// random(50) + 50;
		float noiseVal = noise(frameCount*.025f,seed);
		
		scalar = noiseVal*50 + 50;
		
		pseudoRelativeVec = new PVector( 0, 0, scalar );
		startCrossAbsoluteVec = new PVector( 0, 0, 0 );
		endCrossAbsoluteVec = new PVector( 0, 0, 0 );

		
		startCrossRelativeVec = endRelativeVec.cross(pseudoRelativeVec);
		startCrossRelativeVec.normalize();
		startCrossRelativeVec.mult(scalar);

		endCrossRelativeVec = startCrossRelativeVec.get();
		endCrossRelativeVec.add(endRelativeVec);

	}
	
	
	public void createAbsolutes(){
		startAbsoluteVec = startRelativeVec.get();
		endAbsoluteVec = endRelativeVec.get();
		startCrossAbsoluteVec = startCrossRelativeVec.get();
		endCrossAbsoluteVec = endCrossRelativeVec.get();
		pseudoAbsoluteVec = pseudoRelativeVec.get();
	}
	
	public void updateAbsolutes(PVector offset){
		startAbsoluteVec.add(offset);
		endAbsoluteVec.add(offset);
		startCrossAbsoluteVec.add(offset);
		endCrossAbsoluteVec.add(offset);
		pseudoAbsoluteVec.add(offset);
	}
	
	public void setPreviousOrbit(PVector[] orbiter){
		prevStartOrbit = orbiter;
	}
	public void setNextOrbit(PVector[] orbiter){
		nextStartOrbit = orbiter;
	}
	
	public void parse() {

		PVector startMarginVec = endRelativeVec.get();
		startMarginVec.mult(.2f);
		PVector midVec = endRelativeVec.get();
		midVec.mult(.5f);
		
		PVector endMarginVec = endRelativeVec.get();
		endMarginVec.mult(.8f);
		
		 
		for(int i=0;i<totalLines;i++) {
			angle = 360*PApplet.parseFloat(i)/PApplet.parseFloat(totalLines) +360*(1-PApplet.parseFloat(seed)/PApplet.parseFloat(totalLines)) + frameCount*seed*.025f;

			// public PVector rotateAroundArbitraryAxis(PVector offsetParam,PVector controlParam,PVector axisParam,float angle) {
			// baseOrbit[i] = rotateAroundArbitraryAxis( startRelativeVec, startCrossRelativeVec, zeroVect, angle );
			
			startOrbit[i] = rotateAroundArbitraryAxis( startAbsoluteVec, startCrossAbsoluteVec, endAbsoluteVec, angle );

			startMarginOrbit[i] = startOrbit[i].get();
			startMarginOrbit[i].add(startMarginVec);

			midOrbit[i] = startOrbit[i].get();
			midOrbit[i].add(midVec);
			
			endMarginOrbit[i] = startOrbit[i].get();
			endMarginOrbit[i].add(endMarginVec);

			endOrbit[i] = startOrbit[i].get();
			endOrbit[i].add(endRelativeVec);
			
		}         
	}
	
	public void parseLerps() {
		 
		for(int i=0;i<totalLines;i++) {
			startLerpOrbit[i] = lerp3D(prevStartOrbit[i],startOrbit[i],.5f);
			endLerpOrbit[i] = lerp3D(nextStartOrbit[i],endOrbit[i],.5f);			
		}
	}
	public void drawWireframes() {
		// drawWidthArray(startOrbit,#FFFFFF);
		// drawCirclePointArray(startOrbit,#FFFFFF,51);
		// drawCirclePointArray(endOrbit,#FFFFFF,51);

		drawCirclePointArray(midOrbit,0xff00FF00,51);
		
		drawCirclePointArray(startMarginOrbit,0xffFF00FF,15);
		drawCirclePointArray(endMarginOrbit,0xffFF00FF,15);
		
		drawCirclePointArray(startLerpOrbit,0xffFF0000,255);
		drawCirclePointArray(endLerpOrbit,0xffFF0000,255);
	}
	public void draw() {
		
		PVector curr = new PVector(0,0,0);
		PVector prev = new PVector(0,0,0); 
		
		int lines = endOrbit.length;
		int i;
		for(i=0;i<lines;i++) {
			curr = startOrbit[i].get();
			prev = endOrbit[i].get();
			// drawLine(prev,curr,1,#FFFFFF,51);
			
			noFill();
			stroke(0xffFFFFFF,51);
			strokeWeight(1);
			// curve(x1, y1, z1, x2, y2, z2, x3, y3, z3, x4, y4, z4);
			bezier(
				startAbsoluteVec.x, startAbsoluteVec.y, startAbsoluteVec.z,
				startLerpOrbit[i].x,startLerpOrbit[i].y,startLerpOrbit[i].z,
				startMarginOrbit[i].x,startMarginOrbit[i].y,startMarginOrbit[i].z,
				midOrbit[i].x,midOrbit[i].y,midOrbit[i].z
			);
			
			bezier(
				endAbsoluteVec.x, endAbsoluteVec.y, endAbsoluteVec.z,
				endLerpOrbit[i].x,endLerpOrbit[i].y,endLerpOrbit[i].z,
				endMarginOrbit[i].x,endMarginOrbit[i].y,endMarginOrbit[i].z,
				midOrbit[i].x,midOrbit[i].y,midOrbit[i].z
			);
			
			
		
		}
		
		
	}
	
	public void drawGuides() {
		
		drawPoint(endAbsoluteVec,3,0xffFFFFFF,255);
		drawLine(startAbsoluteVec,endAbsoluteVec,1,0xffFFFFFF,102);

		drawLine(startAbsoluteVec,pseudoAbsoluteVec,1,0xffFFFFFF,102);

		drawPoint(startCrossAbsoluteVec,3,0xffFF00FF,255);
		drawLine(startAbsoluteVec,startCrossAbsoluteVec,1,0xffFF00FF,102);

		drawPoint(endCrossAbsoluteVec,3,0xff00FF00,255);
		drawLine(endAbsoluteVec,endCrossAbsoluteVec,1,0xff00FF00,102);

	}
	public void drawCirclePointArray(PVector[] orbiter, int colorParam,float opacityParam) {
		PVector curr = new PVector(0,0,0);
		int lines = orbiter.length;
		int i;
		float percentage;

		// drawLine(prev,curr,1,#FFFFFF,51);

		for(i=1;i<lines;i++) {
			curr = orbiter[i].get();
			percentage = (PApplet.parseFloat(i)/PApplet.parseFloat(lines));
			drawPoint(curr,1,colorParam,percentage*255);
		}
	}
	public void drawCircleOutlineArray(PVector[] orbiter, int colorParam,float opacityParam) {
		PVector curr = new PVector(0,0,0);
		PVector prev = new PVector(0,0,0);  
		int lines = orbiter.length;
		int i;
		float percentage;

		prev = orbiter[0].get();
		curr = orbiter[lines-1].get();
		// drawLine(prev,curr,1,#FFFFFF,51);

		for(i=1;i<lines;i++) {
			prev = curr.get();
			curr = orbiter[i].get();
			percentage = (PApplet.parseFloat(i)/PApplet.parseFloat(lines));
			drawLine(prev,curr,1,colorParam,percentage*255);
		}
		
	}
	
	// linear interpolation between vectors
	public PVector lerp3D(PVector a, PVector b, float d) {
	  return new PVector(lerp(a.x, b.x, d), lerp(a.y, b.y, d),lerp(a.z, b.z, d)); 
	}

	
	public void drawWidthArray(PVector[] orbiter, int colorParam) {

		PVector curr = new PVector(0,0,0);
		PVector prev = new PVector(0,0,0);  
		int lines = orbiter.length;
		float percentage;
		float colorShifter = sin( ((PApplet.parseFloat(seed) / PApplet.parseFloat(totalSegments) + frameCount*.001f) * TWO_PI))*.5f+.5f;
		strokeWeight(1);
		int interColor = lerpColor(from, to,  colorShifter);

		for(int i=1;i<lines;i++) {
			curr = orbiter[i].get();
			prev = orbiter[i-1].get();

			percentage = (PApplet.parseFloat(i)/PApplet.parseFloat(lines)) * 255;

			noStroke();

			fill(interColor, percentage);
			beginShape(TRIANGLE_FAN);
			vertex(prev.x,prev.y,prev.z);
			vertex(curr.x,curr.y,curr.z);
			vertex(startAbsoluteVec.x,startAbsoluteVec.y,startAbsoluteVec.z);
			vertex(prev.x,prev.y,prev.z);
			endShape();

		}
	}

	public void drawPoint(PVector pointParam, int strokeParam, int colorParam, float opacityParam) {
		strokeWeight(strokeParam);
		stroke(colorParam,opacityParam);
		point( pointParam.x,pointParam.y,pointParam.z);
	}

	public void drawLine(PVector pointAParam,PVector pointBParam, float strokeParam, int colorParam, float opacityParam) {
		strokeWeight(strokeParam);
		stroke(colorParam,opacityParam);
		line( pointAParam.x,pointAParam.y,pointAParam.z,pointBParam.x,pointBParam.y,pointBParam.z);
	}
}

  static public void main(String args[]) {
    PApplet.main(new String[] { "--present", "--bgcolor=#666666", "--stop-color=#cccccc", "Balloon" });
  }
}
