class Segment {

	public PVector offset;
	public PVector axis;
	public PVector control;


	public PVector storedOffset;
	public PVector storedAxis;
	public PVector storedControl;

	public int seed;

	public float angle;

	public int totalLines = 50;

	public PVector[] baseOrbit = new PVector[totalLines];
	public PVector[] startOrbit = new PVector[totalLines];
	public PVector[] startMarginOrbit = new PVector[totalLines];
	public PVector[] endOrbit = new PVector[totalLines];
	public PVector[] endMarginOrbit = new PVector[totalLines];
	public PVector[] midOrbit = new PVector[totalLines];

	public PVector[] prevStartOrbit = new PVector[totalLines];
	public PVector[] nextStartOrbit = new PVector[totalLines];
	
	public PVector[] startLerpOrbit = new PVector[totalLines];
	public PVector[] endLerpOrbit = new PVector[totalLines];

	
	public PVector zeroVect;

	public PVector startRelativeVec;
	public PVector endRelativeVec;
	public PVector startCrossRelativeVec;
	public PVector endCrossRelativeVec;
	public PVector pseudoRelativeVec;

	public PVector startAbsoluteVec;
	public PVector endAbsoluteVec;
	public PVector startCrossAbsoluteVec;
	public PVector endCrossAbsoluteVec;
	public PVector pseudoAbsoluteVec;

	public Segment previousSegment;

	color from = color(255, 0, 102);
	color to = color(0, 102, 153);

	public float scalar = random(50) + 50;

	public Segment() {
		int radius = 25;
		zeroVect	= new PVector( 0,0,0);
		
		axis 	= new PVector( 25, 0, 0 );
		offset 	= new PVector( 0,0,0);
		control = new PVector( 0, 10, 0 );
		
		storedAxis = new PVector( 0, random(25), random(25) );
		storedOffset = offset.get();
		storedControl = control.get();
		
		createRelatives();
		resetRelatives();
		createAbsolutes();
	}

	public void createRelatives(){
		int radius = 300;
		float half_radius = radius*.5;
		startRelativeVec = new PVector( 0, 0, 0 );
		endRelativeVec = new PVector( random(radius)-half_radius, random(radius)-half_radius, random(radius)-half_radius );
		
	}

	public void resetRelatives(){
		
		// scalar = mouseY-height/2;// random(50) + 50;
		float noiseVal = noise(frameCount*.025,seed);
		
		scalar = noiseVal*50 + 50;
		
		pseudoRelativeVec = new PVector( 0, 0, scalar );
		startCrossAbsoluteVec = new PVector( 0, 0, 0 );
		endCrossAbsoluteVec = new PVector( 0, 0, 0 );

		
		startCrossRelativeVec = endRelativeVec.cross(pseudoRelativeVec);
		startCrossRelativeVec.normalize();
		startCrossRelativeVec.mult(scalar);

		endCrossRelativeVec = startCrossRelativeVec.get();
		endCrossRelativeVec.add(endRelativeVec);

	}
	
	
	public void createAbsolutes(){
		startAbsoluteVec = startRelativeVec.get();
		endAbsoluteVec = endRelativeVec.get();
		startCrossAbsoluteVec = startCrossRelativeVec.get();
		endCrossAbsoluteVec = endCrossRelativeVec.get();
		pseudoAbsoluteVec = pseudoRelativeVec.get();
	}
	
	public void updateAbsolutes(PVector offset){
		startAbsoluteVec.add(offset);
		endAbsoluteVec.add(offset);
		startCrossAbsoluteVec.add(offset);
		endCrossAbsoluteVec.add(offset);
		pseudoAbsoluteVec.add(offset);
	}
	
	public void setPreviousOrbit(PVector[] orbiter){
		prevStartOrbit = orbiter;
	}
	public void setNextOrbit(PVector[] orbiter){
		nextStartOrbit = orbiter;
	}
	
	public void parse() {

		PVector startMarginVec = endRelativeVec.get();
		startMarginVec.mult(.2);
		PVector midVec = endRelativeVec.get();
		midVec.mult(.5);
		
		PVector endMarginVec = endRelativeVec.get();
		endMarginVec.mult(.8);
		
		 
		for(int i=0;i<totalLines;i++) {
			angle = 360*float(i)/float(totalLines) +360*(1-float(seed)/float(totalLines)) + frameCount*seed*.025;

			// public PVector rotateAroundArbitraryAxis(PVector offsetParam,PVector controlParam,PVector axisParam,float angle) {
			// baseOrbit[i] = rotateAroundArbitraryAxis( startRelativeVec, startCrossRelativeVec, zeroVect, angle );
			
			startOrbit[i] = rotateAroundArbitraryAxis( startAbsoluteVec, startCrossAbsoluteVec, endAbsoluteVec, angle );

			startMarginOrbit[i] = startOrbit[i].get();
			startMarginOrbit[i].add(startMarginVec);

			midOrbit[i] = startOrbit[i].get();
			midOrbit[i].add(midVec);
			
			endMarginOrbit[i] = startOrbit[i].get();
			endMarginOrbit[i].add(endMarginVec);

			endOrbit[i] = startOrbit[i].get();
			endOrbit[i].add(endRelativeVec);
			
		}         
	}
	
	public void parseLerps() {
		 
		for(int i=0;i<totalLines;i++) {
			startLerpOrbit[i] = lerp3D(prevStartOrbit[i],startOrbit[i],.5);
			endLerpOrbit[i] = lerp3D(nextStartOrbit[i],endOrbit[i],.5);			
		}
	}
	public void drawWireframes() {
		// drawWidthArray(startOrbit,#FFFFFF);
		// drawCirclePointArray(startOrbit,#FFFFFF,51);
		// drawCirclePointArray(endOrbit,#FFFFFF,51);

		drawCirclePointArray(midOrbit,#00FF00,51);
		
		drawCirclePointArray(startMarginOrbit,#FF00FF,15);
		drawCirclePointArray(endMarginOrbit,#FF00FF,15);
		
		drawCirclePointArray(startLerpOrbit,#FF0000,255);
		drawCirclePointArray(endLerpOrbit,#FF0000,255);
	}
	public void draw() {
		
		PVector curr = new PVector(0,0,0);
		PVector prev = new PVector(0,0,0); 
		
		int lines = endOrbit.length;
		int i;
		for(i=0;i<lines;i++) {
			curr = startOrbit[i].get();
			prev = endOrbit[i].get();
			// drawLine(prev,curr,1,#FFFFFF,51);
			
			noFill();
			stroke(#FFFFFF,51);
			strokeWeight(1);
			// curve(x1, y1, z1, x2, y2, z2, x3, y3, z3, x4, y4, z4);
			bezier(
				startAbsoluteVec.x, startAbsoluteVec.y, startAbsoluteVec.z,
				startLerpOrbit[i].x,startLerpOrbit[i].y,startLerpOrbit[i].z,
				startMarginOrbit[i].x,startMarginOrbit[i].y,startMarginOrbit[i].z,
				midOrbit[i].x,midOrbit[i].y,midOrbit[i].z
			);
			
			bezier(
				endAbsoluteVec.x, endAbsoluteVec.y, endAbsoluteVec.z,
				endLerpOrbit[i].x,endLerpOrbit[i].y,endLerpOrbit[i].z,
				endMarginOrbit[i].x,endMarginOrbit[i].y,endMarginOrbit[i].z,
				midOrbit[i].x,midOrbit[i].y,midOrbit[i].z
			);
			
			
		
		}
		
		
	}
	
	public void drawGuides() {
		
		drawPoint(endAbsoluteVec,3,#FFFFFF,255);
		drawLine(startAbsoluteVec,endAbsoluteVec,1,#FFFFFF,102);

		drawLine(startAbsoluteVec,pseudoAbsoluteVec,1,#FFFFFF,102);

		drawPoint(startCrossAbsoluteVec,3,#FF00FF,255);
		drawLine(startAbsoluteVec,startCrossAbsoluteVec,1,#FF00FF,102);

		drawPoint(endCrossAbsoluteVec,3,#00FF00,255);
		drawLine(endAbsoluteVec,endCrossAbsoluteVec,1,#00FF00,102);

	}
	void drawCirclePointArray(PVector[] orbiter, color colorParam,float opacityParam) {
		PVector curr = new PVector(0,0,0);
		int lines = orbiter.length;
		int i;
		float percentage;

		// drawLine(prev,curr,1,#FFFFFF,51);

		for(i=1;i<lines;i++) {
			curr = orbiter[i].get();
			percentage = (float(i)/float(lines));
			drawPoint(curr,1,colorParam,percentage*255);
		}
	}
	void drawCircleOutlineArray(PVector[] orbiter, color colorParam,float opacityParam) {
		PVector curr = new PVector(0,0,0);
		PVector prev = new PVector(0,0,0);  
		int lines = orbiter.length;
		int i;
		float percentage;

		prev = orbiter[0].get();
		curr = orbiter[lines-1].get();
		// drawLine(prev,curr,1,#FFFFFF,51);

		for(i=1;i<lines;i++) {
			prev = curr.get();
			curr = orbiter[i].get();
			percentage = (float(i)/float(lines));
			drawLine(prev,curr,1,colorParam,percentage*255);
		}
		
	}
	
	// linear interpolation between vectors
	PVector lerp3D(PVector a, PVector b, float d) {
	  return new PVector(lerp(a.x, b.x, d), lerp(a.y, b.y, d),lerp(a.z, b.z, d)); 
	}

	
	void drawWidthArray(PVector[] orbiter, color colorParam) {

		PVector curr = new PVector(0,0,0);
		PVector prev = new PVector(0,0,0);  
		int lines = orbiter.length;
		float percentage;
		float colorShifter = sin( ((float(seed) / float(totalSegments) + frameCount*.001) * TWO_PI))*.5+.5;
		strokeWeight(1);
		color interColor = lerpColor(from, to,  colorShifter);

		for(int i=1;i<lines;i++) {
			curr = orbiter[i].get();
			prev = orbiter[i-1].get();

			percentage = (float(i)/float(lines)) * 255;

			noStroke();

			fill(interColor, percentage);
			beginShape(TRIANGLE_FAN);
			vertex(prev.x,prev.y,prev.z);
			vertex(curr.x,curr.y,curr.z);
			vertex(startAbsoluteVec.x,startAbsoluteVec.y,startAbsoluteVec.z);
			vertex(prev.x,prev.y,prev.z);
			endShape();

		}
	}

	public void drawPoint(PVector pointParam, int strokeParam, color colorParam, float opacityParam) {
		strokeWeight(strokeParam);
		stroke(colorParam,opacityParam);
		point( pointParam.x,pointParam.y,pointParam.z);
	}

	public void drawLine(PVector pointAParam,PVector pointBParam, float strokeParam, color colorParam, float opacityParam) {
		strokeWeight(strokeParam);
		stroke(colorParam,opacityParam);
		line( pointAParam.x,pointAParam.y,pointAParam.z,pointBParam.x,pointBParam.y,pointBParam.z);
	}
}

